/****************************************************************
 * music.h
 *---------------------------------------------------------------
 * Copyright (C) 2014-2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in hscc.c
 ****************************************************************/
#ifndef MUSIC
#define MUSIC

#define MUSIC_CHANNELS					HU_CHANNELS

/* Music Header
 -----------------------------*/
#define MUSIC_MAGICWORD					"MUS"
#define MUSIC_HEADER_OFFSET_MAGIC		0
#define MUSIC_HEADER_OFFSET_INST_COUNT	3
#define MUSIC_HEADER_OFFSET_INST_ADDR	4
#define MUSIC_HEADER_OFFSET_INST_ADDRL	MUSIC_HEADER_OFFSET_INST_ADDR + 0
#define MUSIC_HEADER_OFFSET_INST_ADDRH	MUSIC_HEADER_OFFSET_INST_ADDR + 1
#define MUSIC_HEADER_OFFSET_INST_ADDRB	MUSIC_HEADER_OFFSET_INST_ADDR + 2
#define MUSIC_HEADER_OFFSET_INST_ADDRM	MUSIC_HEADER_OFFSET_INST_ADDR + 3
#define MUSIC_HEADER_OFFSET_CHANNELS	8

#define MUSIC_CHANNEL_MAGICWORD			'C'
#define MUSIC_CHANNEL_OFFSET_MAGIC		0
#define MUSIC_CHANNEL_OFFSET_ADDR		1
#define MUSIC_CHANNEL_OFFSET_ADDRL		MUSIC_CHANNEL_OFFSET_ADDR + 0
#define MUSIC_CHANNEL_OFFSET_ADDRH		MUSIC_CHANNEL_OFFSET_ADDR + 1
#define MUSIC_CHANNEL_OFFSET_ADDRB		MUSIC_CHANNEL_OFFSET_ADDR + 2
#define MUSIC_CHANNEL_OFFSET_ADDRM		MUSIC_CHANNEL_OFFSET_ADDR + 3
#define MUSIC_CHANNEL_OFFSET_PRIORITY	5
#define MUSIC_CHANNEL_DESC_LEN			6

#define MUSIC_HEADER_LEN				MUSIC_HEADER_OFFSET_CHANNELS + (MUSIC_CHANNEL_DESC_LEN * 6)

/* Music Script Commands
 -----------------------------*/
#define MUSIC_COMMAND_SET_PRIORITY		0
#define MUSIC_COMMAND_SET_PANNING		1
#define MUSIC_COMMAND_DELAY				2

#define MUSIC_COMMAND_LOOP_START		3
#define MUSIC_COMMAND_LOOP_END			4
#define MUSIC_COMMAND_CALL				5
#define MUSIC_COMMAND_RETURN			6
#define MUSIC_COMMAND_BREAK				7

#define MUSIC_COMMAND_PITCH				8
#define MUSIC_COMMAND_NOTE_ON			9
#define MUSIC_COMMAND_NOTE_OFF			10
#define MUSIC_COMMAND_SAMPLE			11

/* Buffer Setup
 -----------------------------*/
#define MUSIC_PARSER_LINEBUF_LEN		512
#define MUSIC_PARSER_BLOCK_LEN			64
#define MUSIC_PARSER_TAG_LEN			64
#define MUSIC_PARSER_STREAM_LEN			1024

/* Value Ranges
 -----------------------------*/
#define MUSIC_PRIORITY_MIN				0
#define MUSIC_PRIORITY_MAX				255
#define MUSIC_LOOP_MIN					-128
#define MUSIC_LOOP_MAX					127
#define MUSIC_PANNING_MIN				0
#define MUSIC_PANNING_MAX				255
#define MUSIC_SAMPLE_MIN				0
#define MUSIC_SAMPLE_MAX				255
#define MUSIC_WAIT_MIN					0
#define MUSIC_WAIT_MAX					255


/* Block and Tag Containers
 -----------------------------*/
typedef struct MUS_Block
{
	char *pszBlockName;

	char *pacBlockStream;
	unsigned int uiScriptLen,uiBufferLen;
	unsigned int uiOffset;
} MUS_Block;

typedef struct MUS_Tag
{
	char *pszBlockName;
	unsigned int uiOffset;
	
	char *pszSrcFile;
	unsigned int uiLine;
} MUS_Tag;

/****************************************************************
 * Function Defines
 ****************************************************************/
int mus_addNote(MUS_Block *pBlock, char *pszNote);
int mus_parseSASS(FILE *pSASS, char *pszFileName);
int mus_linkSASS(char **pacMusic, unsigned int *puiBinLen);

int mus_addByte(MUS_Block *pBlock, char cByte);
int mus_addBlock(MUS_Block *pBlock);
int mus_addCallTag(char *pszTarget, unsigned int uiOffset,
				   char *pszSrcFile, unsigned int uiLine);
void mus_closeTable(void);
#endif
