/****************************************************************
 * sfx.c
 *---------------------------------------------------------------
 * Copyright (C) 2014-2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in hscc.c
 ****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bytes.h"
#include "useful.h"
#include "hu.h"
#include "sfx.h"
/****************************************************************
 * Globals and Externs
 ****************************************************************/
/* Parsing State
 -----------------------------*/
char *pszSFXFile = NULL;
char *pszParseBuf = NULL;
unsigned int uiSFXLine = 0;
unsigned int uiSFXOffset = 0;


/* Active Blocks
 -----------------------------*/
SFX_Block *paSFXs = NULL;
unsigned int uiSFXs = 0,uiSFXsBufLen = 0;


/****************************************************************
 * Text Tables
 ****************************************************************/
char REM_SFXMode[]=		"SFX Mode...\n";
char REM_SFXDone[]=		"The current SFX binary is : ";
char REM_InstDone[]=	"The current instrument table is : ";
char REM_InstEmpty[]=	"The current instrument table is empty\n";
char REM_EquDone[]=		"Saving SFX equates to : ";

char ERR_NoSFXMem[]=	"ERROR- Out of memory\n";
char ERR_SFXRange[]=	"ERROR- Value out of range\n";
char ERR_SFXBadEqu[]=	"ERROR- Unable to generate equate(s) file\n";
char ERR_SFXBadWrite[]=	"ERROR- Couldn't write to equate(s) file : ";

char ERR_SFXSyntax[]=	"ERROR- Invalid command syntax in : ";
char ERR_SFXGarbage[]=	"ERROR- Garbage at end of line in : ";
char ERR_SFXCom[]=		"ERROR- Invalid command : ";
char ERR_SFXFlow[]=		"ERROR- Flow command outside of script header : ";

char ERR_SFXWave[]=		"ERROR- Waveform has less than 32 samples, saw ";
char ERR_SFXChIns[]=	"ERROR- Instrument uses more than one channel in : ";
char ERR_SFXChMuch[]=	"ERROR- Sound effect has too many channels in : ";
char ERR_SFXChNoise[]=	"ERROR- Sound effect has more than two noise channels : ";
char ERR_SFXNoteOff[]=	"ERROR- Note off assigned for sound effect : ";
char ERR_SFXTuning[]=	"ERROR- Tuning assigned for sound effect : ";

char ERR_SFXBadLabel[]=	"ERROR- Invalid instrument/sfx label in : ";
char ERR_SFXBadPri[]=	"ERROR- Priority out of range in : ";
char ERR_SFXInstPri[]=	"ERROR- Priority assigned to instrument in : ";
char ERR_SFXTwoLabels[]="ERROR- Instrument/sfx already defined with label : ";
char ERR_SFXUnclear[]=	"ERROR- Unclear instrument/sfx syntax in : ";
char ERR_SFXBusy[]=		"ERROR- Unexpected end of instrument/sfx in : ";

char SFX_WhiteSpace[]=	" ,\t\n";

/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----             SFX / INST COMMANDS                  --!! *
 * 
 ****************************************************************/
typedef int (*SFXCom)(SFX_Block *pSFX, SFX_Channel *pChannel);

/* wave,wv waveform
 -----------------------------*/
int sfxCom_wave(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszSubTok;
	int16p16_t stAmp;
	int8p8_t stWrite;

	int iStatus;
	unsigned int uiX;

	iStatus = sfx_addByte(pChannel, SFX_COMMAND_WAVEFORM);
	for(uiX = 0; uiX < 32; uiX++)
	{
		if(NULL == (pszSubTok = strtok(NULL, SFX_WhiteSpace)))
		{
			fprintf(stderr, "%s%u in : %s%u\n",
				ERR_SFXWave, uiX, pszSFXFile, uiSFXLine);
			return 1;
		}
		if(parse16p8(&stAmp, pszSubTok))
			return 1;
		if((0 != stAmp.usPair.usLo) || (stAmp.usPair.usHi > 32))
		{
			fprintf(stderr, "%s", ERR_SFXRange);
			return 1;
		}

		stWrite.sWhole = stAmp.sPair.sHi;
		iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	}

	return iStatus;
}

/* noise,no
 -----------------------------*/
int sfxCom_noise(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	pChannel->iUseNoise = -1;
	return sfx_addByte(pChannel, SFX_COMMAND_LFSR);
}

/* sample,s number
 -----------------------------*/
int sfxCom_sample(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszSubTok;

	int iStatus,iVal;
	char cSample;

	if(NULL == (pszSubTok = strtok(NULL, SFX_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < SFX_SAMPLE_MIN) || (iVal > SFX_SAMPLE_MAX))
	{
		fprintf(stderr, "%s", ERR_SFXRange);
		return 1;
	}

	cSample = iVal & 0xFF;
	iStatus = sfx_addByte(pChannel, SFX_COMMAND_SAMPLE);
	iStatus |= sfx_addByte(pChannel, cSample);
	return iStatus;
}




/* frequency,f offset [adjustment]
 -----------------------------*/
int sfxCom_frequency(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszOffTok,*pszAdjTok;
	int16p16_t stOff,stAdj;
	int8p8_t stWrite;
	int iStatus;

	pszOffTok = strtok(NULL, SFX_WhiteSpace);
	pszAdjTok = strtok(NULL, SFX_WhiteSpace);
	if(NULL == pszOffTok)
		return 1;
	if(parse16p8(&stOff, pszOffTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p8(&stAdj, pszAdjTok))
			return 1;
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = sfx_addByte(pChannel, SFX_COMMAND_FREQUENCY);

	stWrite.usWhole = stOff.usPair.usHi;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cHi);

	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cHi);
	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	return iStatus;
}

/* volume,v amplitude [adjustment]
 -----------------------------*/
int sfxCom_volume(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszVolTok,*pszAdjTok;
	int16p16_t stVol,stAdj;
	int8p8_t stWrite;
	int iStatus;

	pszVolTok = strtok(NULL, SFX_WhiteSpace);
	pszAdjTok = strtok(NULL, SFX_WhiteSpace);
	if(NULL == pszVolTok)
		return 1;
	if(parse16p8(&stVol, pszVolTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p8(&stAdj, pszAdjTok))
			return 1;
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = sfx_addByte(pChannel, SFX_COMMAND_VOLUME);

	stWrite.usWhole = stVol.usPair.usHi;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);

	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= sfx_addByte(pChannel, stWrite.cPair.cLo);
	return iStatus;
}

/* tuning,t value
 -----------------------------*/
int sfxCom_tuning(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszSubTok;

	int iStatus;
	double dTuning;

	if(NULL == (pszSubTok = strtok(NULL, SFX_WhiteSpace)))
		return 1;
	if(0 == sscanf(pszSubTok, "%lf", &dTuning))
		return 1;
	if(dTuning < 0.0)
	{
		fprintf(stderr, "%s", ERR_SFXRange);
		return 1;
	}

	if(pSFX->iAmInstrument)
	{
		pSFX->dTuning = dTuning;
		iStatus = 0;
	}
	else
	{
		fprintf(stderr, "%s%s in : %s,%u\n",
			ERR_SFXTuning, pSFX->pszName, pszSFXFile, uiSFXLine);
		iStatus = 1;
	}
	return iStatus;
}




/* loop,l count
 -----------------------------*/
int sfxCom_loop(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszSubTok;
	int iStatus,iVal;
	char cCount;

	if(NULL == (pszSubTok = strtok(NULL, SFX_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < SFX_LOOP_MIN) || (iVal > SFX_LOOP_MAX))
	{
		fprintf(stderr, "%s", ERR_SFXRange);
		return 1;
	}

	cCount = iVal & 0xFF;
	iStatus = sfx_addByte(pChannel, SFX_COMMAND_LOOP_START);
	iStatus |= sfx_addByte(pChannel, cCount);
	return iStatus;
}

/* endloop,el
 -----------------------------*/
int sfxCom_endloop(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	return sfx_addByte(pChannel, SFX_COMMAND_LOOP_END);
}




/* wait,w ticks
 -----------------------------*/
int sfxCom_wait(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	char *pszSubTok;
	int iStatus,iVal,iSub;
	char cTicks;

	if(NULL == (pszSubTok = strtok(NULL, SFX_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if(iVal < SFX_WAIT_MIN)
	{
		fprintf(stderr, "%s", ERR_SFXRange);
		return 1;
	}

	iStatus = 0;
	while(iVal > 0)
	{
		if(iVal > SFX_WAIT_MAX)
		{
			cTicks = (char)SFX_WAIT_MAX;
			iStatus |= sfx_addByte(pChannel, SFX_COMMAND_DELAY);
			iStatus |= sfx_addByte(pChannel, cTicks);

			iVal -= SFX_WAIT_MAX;
		}
		else
		{
			iSub = iVal % (SFX_WAIT_MAX + 1);
			cTicks = iSub;

			iStatus |= sfx_addByte(pChannel, SFX_COMMAND_DELAY);
			iStatus |= sfx_addByte(pChannel, cTicks);
			iVal -= iSub;
		}
	}
	return iStatus;
}

/* noteoff,n
 -----------------------------*/
int sfxCom_noteoff(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	if(!(pSFX->iAmInstrument))
	{
		fprintf(stderr, "%s%s in : %s,%u\n",
			ERR_SFXNoteOff, pSFX->pszName, pszSFXFile, uiSFXLine);
		return 1;
	}

	pSFX->uiNoteOff = uiSFXOffset;
	return 0;
}

/* end,e
 -----------------------------*/
int sfxCom_end(SFX_Block *pSFX, SFX_Channel *pChannel)
{
	return sfx_addByte(pChannel, SFX_COMMAND_END);
}



/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----              COMMAND REFERENCE                   --!! *
 * 
 ****************************************************************/
char *apszSFXCom_nameLong[]= {
	"wave","noise","sample",
	"frequency","volume","tuning",
	"loop","endloop",
	"wait","rest","noteoff","end",
	NULL
};

char *apszSFXCom_nameShort[]= {
	"wv","no","s",
	"f","v","t",
	"l","el",
	"w","r","n","e",
	NULL
};

SFXCom aSFXCom_addr[]= {
	sfxCom_wave,sfxCom_noise,sfxCom_sample,
	sfxCom_frequency,sfxCom_volume,sfxCom_tuning,
	sfxCom_loop,sfxCom_endloop,
	sfxCom_wait,sfxCom_wait,sfxCom_noteoff,sfxCom_end,
	NULL
};

int aiSFXCom_bodyOnly[]= {
	0,0,0,
	0,0,0,
	-1,-1,
	-1,-1,-1,-1,
	0
};


/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----                BIG BAD PARSER                    --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int sfx_bodySASS(SFX_Block *pSFX, char *pszMainTok, int *piBodyOpen)
 * Handle parsing of a single instrument or sound effect once
 * its label and parsing mode (single or multi channel) have
 * been identified.
 *
 * If any errors are encountered, a nonzero value will be returned.
 ****************************************************************/
int sfx_bodySASS(SFX_Block *pSFX, char *pszMainTok, int *piBodyOpen)
{
	int iStatus;
	int iBodyOpen,iAmInstrument;
	unsigned int uiX,uiY;

	/**
	 * Before actually handling the header and body decoding we
	 * need to make a quick check to see if we're at the end of
	 * one channel's script and moving into the next one's header.
	 *
	 * Essentially:
	 *  -We're not parsing the body
	 *  -The channel has been closed (hit '}')
	 *
	 * If this is the case, we'll set up the next channel's data
	 * stream. Note that the multimode parser in mus_parseSASS is
	 * actually responsible for closing the sound effect based
	 * upon whether it is using single or multichannel syntax.
	 *
	 * However, we will be checking to see if an instrument has
	 * been given more than one channel or a sound effect is
	 * demanding more hardware than actually exists.
	 */
	iStatus = 0;
	iBodyOpen = *piBodyOpen;
	iAmInstrument = pSFX->iAmInstrument;
	uiX = pSFX->uiChannels - 1;
	if(!iBodyOpen && (pSFX->aChannelData[uiX].iAmClosed))
	{
		if(pSFX->iAmInstrument)
		{
			fprintf(stderr, "%s%s,%u\n",
				ERR_SFXChIns, pszSFXFile, uiSFXLine);
			return -1;
		}
		else if((pSFX->uiChannels + 1) > SFX_CHANNELS)
		{
			fprintf(stderr, "%s%s,%u\n",
				ERR_SFXChMuch, pszSFXFile, uiSFXLine);
			return -1;
		}
		pSFX->uiChannels++;
		uiX++;

		pSFX->aChannelData[uiX].pacStream = malloc(sizeof(char) * SFX_PARSER_STREAM_LEN);
		pSFX->aChannelData[uiX].iUseNoise = 0;
		pSFX->aChannelData[uiX].uiScriptLen = 0;
		pSFX->aChannelData[uiX].uiBufferLen = SFX_PARSER_STREAM_LEN;
		if(NULL == pSFX->aChannelData[uiX].pacStream)
		{
			fprintf(stderr, "%s", ERR_NoSFXMem);
			return -1;
		}
	}

	/**
	 * Unless we're opening or closing the channel body with a
	 * curlybrace, we can assume the current main token is
	 * a command and we'll look it up in the command table.
	 */
	if(iBodyOpen && (0 == strcmp(pszMainTok, "}")))
	{
		iBodyOpen = 0;
		pSFX->aChannelData[uiX].iAmClosed = -1;
	}
	else if(!iBodyOpen && (0 == strcmp(pszMainTok, "{")))
	{
		iBodyOpen = -1;
	}
	else
	{
		/**
		 * Not all commands are supported all the time, there's
		 * context involved. For example, a loop can only appear
		 * in the script body and tuning only applies when
		 * parsing in instrument. The former type of cases will
		 * be handled here in dispatch.
		 */
		for(uiY = 0; (NULL != aSFXCom_addr[uiY]); uiY++)
		{
			if((0 == strcmp(pszMainTok, apszSFXCom_nameLong[uiY])) ||
			   (0 == strcmp(pszMainTok, apszSFXCom_nameShort[uiY])))
			{
			   break;
			}
		}

		if(NULL == aSFXCom_addr[uiY])
		{
			fprintf(stderr, "%s%s in : %s,%u\n",
				ERR_SFXCom, pszMainTok, pszSFXFile, uiSFXLine);
			return -1;
		}

		if(!iBodyOpen && aiSFXCom_bodyOnly[uiY])
		{
			fprintf(stderr, "%s%s in : %s,%u\n",
				ERR_SFXFlow, pszMainTok, pszSFXFile, uiSFXLine);
			return -1;
		}

		iStatus = (*aSFXCom_addr[uiY])(pSFX, &(pSFX->aChannelData[uiX]));
		if(iStatus > 0)
		{
			fprintf(stderr, "%s%s,%u\n",
				ERR_SFXSyntax, pszSFXFile, uiSFXLine);
		}
	}

	*piBodyOpen = iBodyOpen;
	return iStatus;
}
/****************************************************************
 * int sfx_parseSASS(FILE *pSASS, char *pszFileName, int iDoSFX)
 * Parse the SASS script in the file pSASS as either a instrument
 * script (iDoSFX == 0) or sound effect script (iDoSFX != 0).
 * As with the music parser, the file's name must be stored in
 * pszFileName in order to generate helpful errors.
 *
 * If any errors are encountered during parsing, a nonzero
 * value will be returned.
 ****************************************************************/
int sfx_parseSASS(FILE *pSASS, char *pszFileName, int iDoSFX)
{
	SFX_Block CurSFX;
	int iNewPri = 0;

	int iAborted = -1,iStatus = 0;
	int iFoundSFX = 0,iMultiOpen = 0,iSingleOpen = 0;
	int iBodyOpen = 0,iNewBody = 0;
	char *pszMainTok = NULL,*pszSubTok = NULL;
	unsigned int uiX,uiY;

	/**
	 * !!-- PARSE SETUP --!!
	 * Set up the globals required for parsing, clear out the
	 * temporary use sound effect container, and alert the
	 * user if we're in sound effects parsing mode.
	 */
	pszSFXFile = pszFileName;

	uiSFXLine = 0;
	uiSFXOffset = 0;

	CurSFX.pszName = NULL;
	CurSFX.uiChannels = 0;
	CurSFX.iAmInstrument = iDoSFX ? 0 : -1;
	for(uiX = 0; uiX < SFX_CHANNELS; uiX++)
	{
		CurSFX.aChannelData[uiX].pacStream = NULL;
	}

	if(iDoSFX)
		fprintf(stdout, "%s", REM_SFXMode);

	paSFXs = malloc(sizeof(SFX_Block) * SFX_PARSER_SFXBUF_LEN);
	if(NULL == paSFXs)
	{
		fprintf(stderr, "%s", ERR_NoSFXMem);
		return -1;
	}
	uiSFXsBufLen = SFX_PARSER_SFXBUF_LEN;


	/**
	 * !!-- PARSE START --!!
	 * Allocate our linebuffer then begin traversing
	 * the SASS source line by line.
	 */
	if(NULL == (pszParseBuf = malloc(sizeof(char) * SFX_PARSER_LINEBUF_LEN)))
	{
		fprintf(stderr, "%s", ERR_NoSFXMem);
		return -1;
	}

	for(uiSFXLine = 1; ; uiSFXLine++)
	{
		/**
		 * Fetch our current line from the source file, stripping
		 * out any ;comments and tokenizing the remaining characters.
		 */
		if(fgets(pszParseBuf, SFX_PARSER_LINEBUF_LEN, pSASS) == NULL)
		{
			iAborted = 0;
			break;
		}
		if(strcut(pszParseBuf, ";", '\0')) break;
		pszMainTok = strtok(pszParseBuf, SFX_WhiteSpace);

		if(iFoundSFX && (NULL != pszMainTok))
		{
			/**
			 * !!-- INSTRUMENT/SFX MULTIMODE PARSER --!!
			 * -----------------------------------------
			 * The original SASS specification defined both
			 * instruments and sound effects as one channel in
			 * length due to the limits of its target (Atari Lynx).
			 *
			 * We're following the new multichannel standard
			 * here, where sound effects may be as many channels
			 * long as there are resources in the target and
			 *
			 * BUT- instruments are still limited to only one
			 * channel width even though they need to follow the
			 * new multichannel syntax.
			 */
			if(iMultiOpen)
			{
				/**
				 * !!-- MULTICHANNEL PARSING --!!
				 * ------------------------------
				 * Woah, new thing: multiple channels are added
				 * to a single sound effect by enclosing all
				 * their headers and bodies in curlybraces which
				 * open right after the sound effect's name.
				 */
				if(!iBodyOpen && (0 == strcmp(pszMainTok, "}")))
				{
					uiY = 0;
					for(uiX = 0; uiX < CurSFX.uiChannels; uiX++)
					{
						if(CurSFX.aChannelData[uiX].iUseNoise)
							uiY++;
					}

					if(uiY > HU_LFSRS)
					{
						fprintf(stderr, "%s%s in : %s,%u\n",
							ERR_SFXChNoise, CurSFX.pszName, pszFileName, uiSFXLine);
						break;
					}

					if(sfx_addSFX(&CurSFX))
						break;

					CurSFX.pszName = NULL;
					CurSFX.dTuning = -1.0;
					CurSFX.uiChannels = 0;
					iFoundSFX = 0;
				}
				else
				{
					if(sfx_bodySASS(&CurSFX, pszMainTok, &iBodyOpen))
						break;
				}
			}
			else
			{
				/**
				 * Since the user could be writing either a single or
				 * multichannel definition here, we'll make the call
				 * based upon if we see '{' before any valid commands.
				 *
				 * If we see an opening curlybrace first, then this
				 * is (very) likely to be a multichannel sound effect
				 * definition. If it's one of our valid channel header
				 * commands, then we're probably decoding something
				 * in the old single channel syntax.
				 */
				if(iSingleOpen || (0 != strcmp(pszMainTok, "{")))
				{
					/**
					 * !!-- SINGLE CHANNEL PARSING --!!
					 * --------------------------------
					 * Just like in the old SASSy days, only the sound
					 * effect or instrument body is curlybraced.
					 */
					if(sfx_bodySASS(&CurSFX, pszMainTok, &iBodyOpen))
						break;

					iSingleOpen = !(CurSFX.aChannelData[0].iAmClosed);
					if(!iSingleOpen)
					{
						if(sfx_addSFX(&CurSFX))
							break;

						CurSFX.pszName = NULL;
						CurSFX.dTuning = 32.0;
						CurSFX.uiChannels = 0;
						iFoundSFX = 0;
					}
				}
				else
				{
					if(!iSingleOpen)
					{
						iMultiOpen = -1;
					}
					else
					{
						fprintf(stderr, "%s%s,%u\n",
							ERR_SFXUnclear, pszFileName, uiSFXLine);
						break;
					}
				}
			}
		}
		else if(NULL != pszMainTok)
		{
			/**
			 * !!-- INSTRMENT / SFX DETECTION --!!
			 * -----------------------------------
			 * We're not currently parsing an instrument or sound
			 * effect, so we need to look for the start of one with
			 * the format name [priority].
			 */
			pszSubTok = strtok(NULL, SFX_WhiteSpace);
			if(NULL != pszSubTok)
			{
				/**
				 * Priorities should just be an integer value after
				 * the sound effect name. However, they're not allowed
				 * for instruments since their priorities are taken
				 * from the currently decoding music track.
				 */
				if(1 == sscanf(pszSubTok, "%d", &iNewPri))
				{
					if((iNewPri < SFX_PRIORITY_MIN) || (iNewPri > SFX_PRIORITY_MAX))
					{
						fprintf(stderr, "%s%s,%u\n",
							ERR_SFXBadPri, pszFileName, uiSFXLine);
						break;
					}

					if(!iDoSFX)
					{
						fprintf(stderr, "%s%s,%u\n",
							ERR_SFXInstPri, pszFileName, uiSFXLine);
						break;
					}
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n",
						ERR_SFXBadLabel, pszFileName, uiSFXLine);
					break;
				}
			}

			/**
			 * Check to see if the instrument or sound effect
			 * already exists, setting up the temporary use
			 * sound effect if not. We'll assume the new sound
			 * effect or instrument to have one channel until
			 * indicated otherwise.
			 */
			if(sfx_find(pszMainTok) >= 0)
			{
				fprintf(stderr, "%s%s in : %s,%u\n",
					ERR_SFXTwoLabels, pszMainTok, pszFileName, uiSFXLine);
				break;
			}

			CurSFX.uiPriority = iNewPri;
			CurSFX.dTuning = -1.0;
			CurSFX.uiChannels = 1;
			for(uiX = 0; uiX < SFX_CHANNELS; uiX++)
			{
				CurSFX.aChannelData[uiX].iAmClosed = 0;
				CurSFX.aChannelData[uiX].iUseNoise = 0;
				CurSFX.aChannelData[uiX].pacStream = NULL;
				CurSFX.aChannelData[uiX].uiBufferLen = 0;
				CurSFX.aChannelData[uiX].uiScriptLen = 0;
			}

			CurSFX.pszName = malloc(sizeof(char) * (strlen(pszMainTok) + 1));
			CurSFX.aChannelData[0].pacStream = malloc(sizeof(char) * SFX_PARSER_STREAM_LEN);
			CurSFX.aChannelData[0].iUseNoise = 0;
			CurSFX.aChannelData[0].uiScriptLen = 0;
			CurSFX.aChannelData[0].uiBufferLen = SFX_PARSER_STREAM_LEN;
			if((NULL == CurSFX.pszName) || (NULL == CurSFX.aChannelData[0].pacStream))
			{
				fprintf(stderr, "%s", ERR_NoSFXMem);
				break;
			}
			strcpy(CurSFX.pszName, pszMainTok);

			iFoundSFX = -1;
			iMultiOpen = 0;
			iSingleOpen = 0;
			iBodyOpen = 0;
		}

		/**
		 * Regardless of mode, check for garbage at the end
		 * of the current parsing line before advancing.
		 */
		pszSubTok = strtok(NULL, SFX_WhiteSpace);
		if(NULL != pszSubTok)
		{
			fprintf(stderr, "%s%s,%u\n",
				ERR_SFXGarbage, pszFileName, uiSFXLine);
			break;
		}
	}

	/**
	 * Check to see if we bailed during the parsing of an
	 * instrument or sound effect and then free out our line
	 * buffer and any dynamic portions of the temporary sound
	 * effect which may be in use.
	 */
	if(iFoundSFX)
	{
		fprintf(stderr, "%s%s,%u\n",
			ERR_SFXBusy, pszFileName, uiSFXLine);
	}

	if(NULL != pszParseBuf) free(pszParseBuf);
	pszParseBuf = NULL;

	if(NULL != CurSFX.pszName) free(CurSFX.pszName);
	CurSFX.pszName = NULL;
	for(uiX = 0; uiX < CurSFX.uiChannels; uiX++)
	{
		if(NULL != CurSFX.aChannelData[uiX].pacStream)
		{
			free(CurSFX.aChannelData[uiX].pacStream);
			CurSFX.aChannelData[uiX].pacStream = NULL;
		}
	}

	return (iAborted | iFoundSFX);
}




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----            SFX-SPECIFIC FUNCTIONS                --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int sfx_buildSFX(char **ppacSFX, unsigned int *puiBinLen)
 * Build a sound effects block from the parsed data.
 * Returns nonzero if anything went wrong.
 ****************************************************************/
int sfx_buildSFX(char **ppacSFX, unsigned int *puiBinLen)
{
	char *pacSFX;
	unsigned int uiBinLen;
	unsigned int uiScriptBase,uiScriptLen;

	HuAddress ResAddr;
	unsigned int uiX,uiY,uiZ,uiT,uiS;

	/**
	 * Sound effect binaries are arranged in
	 * the followng structure:
	 *	-Header and Directory Entries
	 *		"SFX" and the total sound effects count followed
	 *		by relative HuAddresses (4-Bytes Each) specifying
	 *		the starting offset for each instrument script.
	 *	-Sound Effects
	 *		The channel scripts for the sound effects
	 *		prepended with a variable with header indicating
	 *		their individual requirements.
	 *
	 * All relative addresses are caculated relative to
	 * the sound effect package base. But first we need
	 * to figure out how much space this is all
	 * going to take up.
	 */
	uiScriptLen = 0;
	uiScriptBase = SFX_PACK_LEN + (SFX_PACK_DIR_LEN * uiSFXs);
	for(uiX = 0; uiX < uiSFXs; uiX++)
	{
		uiScriptLen += SFX_HEADER_LEN;
		for(uiY = 0; uiY < paSFXs[uiX].uiChannels; uiY++)
		{
			uiScriptLen += SFX_CHANNEL_DESC_LEN;
			uiScriptLen += paSFXs[uiX].aChannelData[uiY].uiScriptLen;
		}
	}

	uiBinLen = uiScriptBase + uiScriptLen;
	fprintf(stdout, "A total of %u SFX were found\n", uiSFXs);
	fprintf(stdout, "%s%u bytes\n", REM_SFXDone, uiBinLen);
	pacSFX = malloc(sizeof(char) * uiBinLen);
	if(NULL == pacSFX)
	{
		fprintf(stderr, "%s", ERR_NoSFXMem);
		return -1;
	}

	strcpy((pacSFX + SFX_PACK_OFFSET_MAGIC), SFX_MAGICWORD);
	pacSFX[SFX_PACK_OFFSET_COUNT] = uiSFXs;
	uiY = SFX_PACK_OFFSET_DIR;
	uiZ = uiScriptBase;
	for(uiX = 0; uiX < uiSFXs; uiX++)
	{
		ResAddr = hu_calcAddr(uiZ);
		pacSFX[uiY + HU_LO] = ResAddr.ucLow;
		pacSFX[uiY + HU_HI] = ResAddr.ucHigh;
		pacSFX[uiY + HU_BANK] = ResAddr.ucBank;
		pacSFX[uiY + HU_MAP] = ResAddr.ucMap;
		uiY += SFX_PACK_DIR_LEN;

		pacSFX[uiZ + SFX_HEADER_OFFSET_PRI] = paSFXs[uiX].uiPriority;
		pacSFX[uiZ + SFX_HEADER_OFFSET_COUNT] = paSFXs[uiX].uiChannels;
		uiZ += SFX_HEADER_LEN;
		uiT = uiZ + (SFX_CHANNEL_DESC_LEN * paSFXs[uiX].uiChannels);
		for(uiS = 0; uiS < paSFXs[uiX].uiChannels; uiS++)
		{
			ResAddr = hu_calcAddr(uiT);
			pacSFX[uiZ + SFX_CHANNEL_OFFSET_REQS] =
				paSFXs[uiX].aChannelData[uiS].iUseNoise ? -1 : 0;
			pacSFX[uiZ + SFX_CHANNEL_OFFSET_ADDRL] = ResAddr.ucLow;
			pacSFX[uiZ + SFX_CHANNEL_OFFSET_ADDRH] = ResAddr.ucHigh;
			pacSFX[uiZ + SFX_CHANNEL_OFFSET_ADDRB] = ResAddr.ucBank;
			pacSFX[uiZ + SFX_CHANNEL_OFFSET_ADDRM] = ResAddr.ucMap;

			memcpy(pacSFX + uiT, paSFXs[uiX].aChannelData[uiS].pacStream,
				(sizeof(char) * paSFXs[uiX].aChannelData[uiS].uiScriptLen));
			uiT += paSFXs[uiX].aChannelData[uiS].uiScriptLen;
			uiZ += SFX_CHANNEL_DESC_LEN;
		}
		uiZ = uiT;
	}

	*ppacSFX = pacSFX;
	*puiBinLen = uiBinLen;
	return 0;
}
/****************************************************************
 * int sfx_writeDefs(char *pszSrcName)
 * Create a sound effects equates file for the current sound
 * effects set which may be included with whatever the
 * assembler of the week is.
 * Returns nonzero if anything went wrong.
 ****************************************************************/
int sfx_writeDefs(char *pszSrcName)
{
	FILE *pEqus = NULL;
	char *pszEquName = NULL;
	unsigned int uiX;

	if(changeExt(pszSrcName, &pszEquName, ".equ", "<>/\\:"))
	{
		fprintf(stderr, "%s", ERR_SFXBadEqu);
		return -1;
	}

	if(NULL == (pEqus = fopen(pszEquName, "w")))
	{
		fprintf(stderr, "%s%s\n", ERR_SFXBadWrite, pszEquName);
		return -1;
	}

	fprintf(pEqus, "; PCE-SASS Autogenerated equates for %s\n", pszSrcName);
	fprintf(pEqus, "; You may include these in your WLA-DX project.\n\n");

	fprintf(pEqus, "; SOUND EFFECT IDs : \n");
	fprintf(pEqus, ";--------------------\n");
	for(uiX = 0; uiX < uiSFXs; uiX++)
		fprintf(pEqus, ".define SFX_%s\t\t%u\n", paSFXs[uiX].pszName, uiX);

	fprintf(pEqus, "\n");
	fprintf(pEqus, "; SOUND EFFECT PRIORITIES : \n");
	fprintf(pEqus, ";---------------------------\n");
	for(uiX = 0; uiX < uiSFXs; uiX++)
		fprintf(pEqus, ".define PRI_%s\t\t%u\n", paSFXs[uiX].pszName, paSFXs[uiX].uiPriority);

	fprintf(stdout, "%s%s\n", REM_EquDone, pszEquName);
	fclose(pEqus);
	free(pszEquName);
	return 0;
}




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----         INSTRUMENT-SPECIFIC FUNCTIONS            --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int sfx_getInstInfo(unsigned int uiInstrument, int *piUseNoise,
 *					   double *pdTuning)
 * Fetch the waveform requirements and tuning for the given
 * instrument number (which may be found with sfx_findInst).
 * Returns nonzero if anything went wrong.
 ****************************************************************/
int sfx_getInstInfo(unsigned int uiInstrument, int *piUseNoise,
					double *pdTuning)
{
	if(uiInstrument < uiSFXs)
	{
		*piUseNoise = paSFXs[uiInstrument].aChannelData[0].iUseNoise;
		*pdTuning = paSFXs[uiInstrument].dTuning;
		return 0;
	}
	return -1;
}
/****************************************************************
 * int sfx_buildInst(char **ppacInst, unsigned int *puiBinLen,
 *					 unsigned int *puiNumInst)
 * Build an instrument block for inclusion in a music track with
 * its data automatically malloc'd in *pacInst and length in *puiBinLen.
 * Returns nonzero if anything went wrong.
 ****************************************************************/
int sfx_buildInst(char **ppacInst, unsigned int *puiBinLen,
				  unsigned int *puiNumInst)
{
	char *pacInst;
	unsigned int uiBinLen;
	unsigned int uiScriptBase,uiScriptLen;

	HuAddress ResAddr;
	unsigned int uiX,uiY,uiZ;

	/**
	 * It is completely possible to create a piece of music
	 * with no instruments whatsoever (samples only), in which
	 * case we'll just create a binary of length zero.
	 */
	if(0 == uiSFXs)
	{
		fprintf(stdout, "%s", REM_InstEmpty);

		*ppacInst = NULL;
		*puiBinLen = 0;
		*puiNumInst = 0;
		return 0;
	}

	/**
	 * However, if there are actually instruments we need to
	 * arrange them in the following structure:
	 *	-Directory Entries
	 *		Relative HuAddresses (4-Bytes Each) specifying the
	 *		starting offset for each instrument script.
	 *	-Script Data
	 *		Referenced by the directories above, and prepended
	 *		with another relative address specifying where
	 *		the note off script starts.
	 *
	 * All relative addresses are calculated relative to the
	 * instrument data block start for both our convenience
	 * and the driver's.
	 *
	 * Don't forget that all instruments should only have one
	 * channel of script data, so we'll just be pulling from there.
	 */
	uiScriptLen = 0;
	uiScriptBase = uiSFXs * SFX_PACK_DIR_LEN;
	for(uiX = 0; uiX < uiSFXs; uiX++)
	{
		uiScriptLen += SFX_NOTEOFF_LEN;
		uiScriptLen += paSFXs[uiX].aChannelData[0].uiScriptLen;
	}
	uiBinLen = uiScriptBase + uiScriptLen;

	fprintf(stdout, "A total of %u instruments were found\n", uiSFXs);
	fprintf(stdout, "%s%u bytes\n", REM_InstDone, uiBinLen);
	pacInst = malloc(sizeof(char) * uiBinLen);
	if(NULL == pacInst)
	{
		fprintf(stderr, "%s", ERR_NoSFXMem);
		return -1;
	}

	uiY = 0;
	uiZ = uiScriptBase;
	for(uiX = 0; uiX < uiSFXs; uiX++)
	{
		ResAddr = hu_calcAddr(uiZ);
		pacInst[uiY + HU_LO] = ResAddr.ucLow;
		pacInst[uiY + HU_HI] = ResAddr.ucHigh;
		pacInst[uiY + HU_BANK] = ResAddr.ucBank;
		pacInst[uiY + HU_MAP] = ResAddr.ucMap;
		uiY += SFX_PACK_DIR_LEN;

		ResAddr = hu_calcAddr(uiScriptBase + paSFXs[uiX].uiNoteOff
			+ (SFX_NOTEOFF_LEN * (uiX + 1)));
		pacInst[uiZ + HU_LO] = ResAddr.ucLow;
		pacInst[uiZ + HU_HI] = ResAddr.ucHigh;
		pacInst[uiZ + HU_BANK] = ResAddr.ucBank;
		pacInst[uiZ + HU_MAP] = ResAddr.ucMap;
		uiZ += SFX_NOTEOFF_LEN;

		memcpy(pacInst + uiZ, paSFXs[uiX].aChannelData[0].pacStream,
			(sizeof(char) * paSFXs[uiX].aChannelData[0].uiScriptLen));
		uiZ += paSFXs[uiX].aChannelData[0].uiScriptLen;
	}

	*ppacInst = pacInst;
	*puiBinLen = uiBinLen;
	*puiNumInst = uiSFXs;
	return 0;
}




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----               DATA MANAGEMENT                    --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int sfx_addByte(MUS_Block *pBlock, char cByte)
 * Add the given byte to this sound effect channel's data stream,
 * returning nonzero if any errors occur.
 ****************************************************************/
int sfx_addByte(SFX_Channel *pChannel, char cByte)
{
	if(tealloc(&(pChannel->pacStream), &(pChannel->uiBufferLen),
		(pChannel->uiScriptLen + 1), sizeof(char))) return -1;

	pChannel->pacStream[pChannel->uiScriptLen] = cByte;
	pChannel->uiScriptLen++;
	uiSFXOffset++;
	return 0;
}
/****************************************************************
 * int sfx_addSFX(SFX_Block *pSFX)
 * Adds the sound effect *pSFX to the current sound effect table,
 * returning nonzero if any errors occur.
 ****************************************************************/
int sfx_addSFX(SFX_Block *pSFX)
{
	if(tealloc(&(paSFXs), &uiSFXsBufLen, (uiSFXs + 1),
		sizeof(SFX_Block))) return -1;

	paSFXs[uiSFXs] = *pSFX;
	uiSFXs++;
	return 0;
}
/****************************************************************
 * int sfx_find(char *pszName)
 * Attempt to locate the identification number of the instrument
 * or sound effect whose name is in *pszName. Returns a negative
 * value if the entry could not be found and the instrument or
 * sound effect's ID otherwise.
 ****************************************************************/
int sfx_find(char *pszName)
{
	unsigned int uiX;

	if((NULL == paSFXs) || (NULL == pszName))
		return -1;

	for(uiX = 0; uiX < uiSFXs; uiX++)
	{
		if(0 == strcmp(paSFXs[uiX].pszName, pszName))
			return uiX;
	}

	return -1;
}
/****************************************************************
 * void sfx_closeTable(void)
 * Close the currently active sound effects table. All memory will
 * be released back into the wild and the table's base entries
 * retagged as NULL.
 ****************************************************************/
void sfx_closeTable(void)
{
	unsigned int uiX,uiY;

	if(NULL != paSFXs)
	{
		for(uiX = 0; uiX < uiSFXs; uiX++)
		{
			for(uiY = 0; uiY < paSFXs[uiX].uiChannels; uiY++)
			{
				if(NULL != paSFXs[uiX].aChannelData[uiY].pacStream)
				{
					free(paSFXs[uiX].aChannelData[uiY].pacStream);
				}
			}

			if(NULL != paSFXs[uiX].pszName)
			{
				free(paSFXs[uiX].pszName);
			}
		}

		free(paSFXs);
	}

	paSFXs = NULL;
	uiSFXs = 0;
	uiSFXsBufLen = 0;
}
