;****************************************************************
; Sound Test Status Drawing Subroutines				*
;****************************************************************
; Copyright (C) 2014-2015 Osman Celimli
; For conditions of distribution and use, see copyright notice in HuListen.asm
;----------------------------------------------------------------
;

;****************************************************************
; Audition_DrawState:						*
; Update the state display for the given channel in X.		*
;****************************************************************
.section "Audition_DrawState" force
Audition_DrawState:
	LDA HuSample_Playing,X
	beq _drawNormal

_drawPCM				; ---- PCM CHANNEL ----
	LDA #PSG_CONTROL_VOLMASK	; <- PCM Streams are always played
	LDY #12				; with maximum volume, just varying
	jsr _OffsetHex			; panning.

	LDA HuSample_Priority,X		; <- Make the sweeping assumption that
	CMP #$FF			; any sample with a priority of 255 was
	beq _isSimplePCM		; likely started by the simple playback
					; routine and forced to center panning.
	LDA HuSound_Panning,X
_isSimplePCM
	LDY #22
	jsr _OffsetHex

	CLA
	LDY #33
	jsr _OffsetHex
	CLA
	LDY #35
	jsr _OffsetHex

	LDA HuSample_Priority,X
	LDY #76
	jsr _OffsetHex

	bra _drawShared


_drawNormal				; ---- LESS INTERESTING CHANNEL ----
	LDA HuSound_Vol_Lo,X
	AND #PSG_CONTROL_VOLMASK
	LDY #12
	jsr _OffsetHex

	LDA HuSound_Panning,X
	LDY #22
	jsr _OffsetHex

	LDA HuSound_FreqFinal_Hi,X
	AND #$0F
	LDY #33
	jsr _OffsetHex
	LDA HuSound_FreqFinal_Lo,X
	LDY #35
	jsr _OffsetHex

	LDA HuSound_Priority,X
	LDY #76
	jsr _OffsetHex


_drawShared				; ---- SHARED FINISHING ROUTINES ----
	TXA				; Don't bother rendering and transferring
	EOR FrameCounter		; every channel's waveform every single
	AND #1				; frame, changing them that often is just
	bne _startRender		; excessive. We'll alternate evens/odds.

	jmp _setMode
_startRender
	jsr _waveRender			; Soft-render this channel's currently
					; selected waveform, then set up its
					; transfer to VRAM...

	LDA #(VDC_TRANSFER_I1 | VDC_TRANSFER_1BBG4)
	STA VDAT_Type			; We'll be using 1bpp -> 4bpp expansion

	clc
	LDA #<Audition_Waves
	ADC.w _waveOffsLo,X
	STA VDAT_Src + LO
	LDA #>Audition_Waves
	ADC.w _waveOffsHi,X
	STA VDAT_Src + HI
	STZ VDAT_Src + BANK
	STZ VDAT_Src + MCMAP

	LDA #(8 * 8)
	STA VDAT_Size + LO
	STZ VDAT_Size + HI

	clc
	LDA #<VMEM_USERBASE		; Since these soft tiles are the only
	ADC.w _waveDestsLo,X		; custom things in the sound test screen,
	STA VDAT_Dest + LO		; we'll just put them at the base of
	LDA #>VMEM_USERBASE		; user VRAM.
	ADC.w _waveDestsHi,X
	STA VDAT_Dest + HI
	jsr VDC_MakeVREQ
	jmp _setMode


; Text Tables
;-------------
_channelBaseLo	.db $C0,$80,$40,$00,$C0,$80
_channelBaseHi	.db $00,$01,$02,$03,$03,$04

_modeOff	.db "Idle ",0
_modeWave	.db "Wave ",0
_modeNoise	.db "Noise",0
_modeSample	.db "PCM  ",0
_modeBlocked	.db "Halt ",0




; Waveform Renderer
;----------------------------------------------------------------
_waveRender
	jsr _waveClear			; Preclear the waveform memory using
					; _waveClear so we can just walk through
	clc				; and set the appropriate bits.
	LDA.w HuSound_WaveOffs,X
	ADC #<HuSound_WaveMem		; We want the waves to render in eight 1bpp tiles:
	STA GLO_PointerC + LO		; |   /\   ->  ACEG  | 
	LDA #>HuSound_WaveMem		; |  /  \  ->  BDFH  | 
	ADC #0
	STA GLO_PointerC + HI		; GLO_PointerC = Base Address of our waveform

	clc
	LDA #<Audition_Waves
	ADC.w _waveOffsLo,X
	STA GLO_PointerD + LO
	LDA #>Audition_Waves
	ADC.w _waveOffsHi,X
	STA GLO_PointerD + HI		; GLO_PointerD = Base Address of the soft-tiles

	PHX
	CLY
_waveWalk
	LDA (GLO_PointerC),Y		; The particular horizontal line of pixels
	LSR				; we want to address is:
	AND #$0F			; Ampltiude[Index] + (Index / 8) * 16
	EOR #$0F			; <- Inversion is required otherwise larger
	clc				; amplitudes will grow down, which is a bit
	ADC _tileOffs,Y			; contrary to what is usually expected.
	TAX

_waveFill
	LDA _ampMask,Y			; Then we can select the pixel within the
	SXY				; line just based upon its index.
	ORA (GLO_PointerD),Y
	STA (GLO_PointerD),Y
	SXY

	INX
	TXA				; We'll then continue to fill the pixels
	CMP _tileOffs + 8,Y		; BELOW the amplitude which should make
	bne _waveFill			; the wave read a little more clearly.


	INY
	CPY #32
	bne _waveWalk
	PLX
	rts


_waveClear
	clc
	LDA #<Audition_Waves
	ADC.w _waveOffsLo,X
	STA GLO_PointerD + LO
	LDA #>Audition_Waves
	ADC.w _waveOffsHi,X
	STA GLO_PointerD + HI

	LDY #((8 * 8) - 1)
	CLA
_waveClear_writeLoop
	STA (GLO_PointerD),Y
	DEY
	bpl _waveClear_writeLoop
	rts

_waveOffsLo	.db <(8 * 8 * 0),<(8 * 8 * 1),<(8 * 8 * 2)
		.db <(8 * 8 * 3),<(8 * 8 * 4),<(8 * 8 * 5)
_waveOffsHi	.db >(8 * 8 * 0),>(8 * 8 * 1),>(8 * 8 * 2)
		.db >(8 * 8 * 3),>(8 * 8 * 4),>(8 * 8 * 5)

_waveDestsLo	.db <(16 * 8 * 0),<(16 * 8 * 1),<(16 * 8 * 2)
		.db <(16 * 8 * 3),<(16 * 8 * 4),<(16 * 8 * 5)
_waveDestsHi	.db >(16 * 8 * 0),>(16 * 8 * 1),>(16 * 8 * 2)
		.db >(16 * 8 * 3),>(16 * 8 * 4),>(16 * 8 * 5)

_tileOffs	.db 0,0,0,0,0,0,0,0
		.db 16,16,16,16,16,16,16,16
		.db 32,32,32,32,32,32,32,32
		.db 48,48,48,48,48,48,48,48
		.db 64,64,64,64,64,64,64,64

_ampMask	.db $80,$40,$20,$10,$08,$04,$02,$01
		.db $80,$40,$20,$10,$08,$04,$02,$01
		.db $80,$40,$20,$10,$08,$04,$02,$01
		.db $80,$40,$20,$10,$08,$04,$02,$01


; Typewriter Tools
;----------------------------------------------------------------
_OffsetDraw
	TYA
	clc
	ADC.w _channelBaseLo,X
	STA DBG_TextDest + LO
	LDA.w _channelBaseHi,X
	ADC #0
	STA DBG_TextDest + HI
	jmp DBG_Print

_OffsetHex
	PHA
	TYA
	clc
	ADC.w _channelBaseLo,X
	STA DBG_TextDest + LO
	LDA.w _channelBaseHi,X
	ADC #0
	STA DBG_TextDest + HI
	PLA
	jmp DBG_PrintHex


; Stamp MODE
;------------
_setMode
	LDY #87
	LDA HuSample_Playing,X
	beq _notSampling

	LDA #<_modeSample
	STA DBG_TextSrc + LO
	LDA #>_modeSample
	STA DBG_TextSrc + HI
	jmp _OffsetDraw

_notSampling
	LDA HuSound_noWriteback,X
	beq _notBlocked

	LDA #<_modeBlocked
	STA DBG_TextSrc + LO
	LDA #>_modeBlocked
	STA DBG_TextSrc + HI
	jmp _OffsetDraw

_notBlocked
	LDA HuSound_Priority,X
	bne _isActive

	LDA #<_modeOff
	STA DBG_TextSrc + LO
	LDA #>_modeOff
	STA DBG_TextSrc + HI
	jmp _OffsetDraw

_isActive
	LDA HuSound_WaveMode,X
	bne _isNoise

	LDA #<_modeWave
	STA DBG_TextSrc + LO
	LDA #>_modeWave
	STA DBG_TextSrc + HI
	jmp _OffsetDraw

_isNoise
	LDA #<_modeNoise
	STA DBG_TextSrc + LO
	LDA #>_modeNoise
	STA DBG_TextSrc + HI
	jmp _OffsetDraw
.ends
